/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "hatziSlipGenNewtonianRelaxationFvPatchField.H"
#include "addToRunTimeSelectionTable.H"
#include "fvPatchFieldMapper.H"
#include "volFields.H"
#include "surfaceFields.H"
#include "symmTransformField.H"
#include "singlePhaseTransportModel.H"
#include "turbulenceModel.H"
#include "fieldTypes.H"
#include "incompressible/turbulenceModel/turbulenceModel.H"
#include "RASModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

hatziSlipGenNewtonianRelaxationFvPatchVectorField::hatziSlipGenNewtonianRelaxationFvPatchVectorField
(
    const fvPatch& p,
    const DimensionedField<vector, volMesh>& iF
)
:
    fixedValueFvPatchField<vector>(p, iF),
    rho_(0),
    kh2_(0),
    kh1_(0),
    relaxationFactor_(0)
{}


hatziSlipGenNewtonianRelaxationFvPatchVectorField::hatziSlipGenNewtonianRelaxationFvPatchVectorField
(
    const hatziSlipGenNewtonianRelaxationFvPatchVectorField& ptf,
    const fvPatch& p,
    const DimensionedField<vector, volMesh>& iF,
    const fvPatchFieldMapper& mapper
)
:
    fixedValueFvPatchField<vector>(ptf, p, iF, mapper),
    rho_(ptf.rho_),
    kh2_(ptf.kh2_),
    kh1_(ptf.kh1_),
    relaxationFactor_(ptf.relaxationFactor_)
{}


hatziSlipGenNewtonianRelaxationFvPatchVectorField::hatziSlipGenNewtonianRelaxationFvPatchVectorField
(
    const fvPatch& p,
    const DimensionedField<vector, volMesh>& iF,
    const dictionary& dict
)
:
    fixedValueFvPatchField<vector>(p, iF),
    rho_(readScalar(dict.lookup("rho"))),
    kh2_(readScalar(dict.lookup("kh2"))),
    kh1_(readScalar(dict.lookup("kh1"))),
    relaxationFactor_(readScalar(dict.lookup("relaxationFactor")))
{
    if (dict.found("value"))
    {
        fvPatchField<vector>::operator=
        (
            vectorField("value", dict, p.size())
        );
    }
    else
    {
        // Evaluate the wall velocity
        updateCoeffs();
    }
}


hatziSlipGenNewtonianRelaxationFvPatchVectorField::hatziSlipGenNewtonianRelaxationFvPatchVectorField
(
    const hatziSlipGenNewtonianRelaxationFvPatchVectorField& fcvpvf
)
:
    fixedValueFvPatchField<vector>(fcvpvf),
    rho_(fcvpvf.rho_),
    kh2_(fcvpvf.kh2_),
    kh1_(fcvpvf.kh1_),
    relaxationFactor_(fcvpvf.relaxationFactor_)
{}


hatziSlipGenNewtonianRelaxationFvPatchVectorField::hatziSlipGenNewtonianRelaxationFvPatchVectorField
(
    const hatziSlipGenNewtonianRelaxationFvPatchVectorField& fcvpvf,
    const DimensionedField<vector, volMesh>& iF
)
:
    fixedValueFvPatchField<vector>(fcvpvf, iF),
    rho_(fcvpvf.rho_),
    kh2_(fcvpvf.kh2_),
    kh1_(fcvpvf.kh1_),
    relaxationFactor_(fcvpvf.relaxationFactor_)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void hatziSlipGenNewtonianRelaxationFvPatchVectorField::updateCoeffs()
{
    if (updated())
    {
        return;
    }
    
    //face normal vector
    vectorField nHat = this->patch().nf();  
    //Info << "nhat" << nHat <<endl;
    // Du/dn
    vectorField  gradient = this->snGrad();
    vectorField gradient1 = this->snGrad();    
    // only tangential components
    gradient = transform(I - sqr(nHat), gradient);
    
    // gradient Direction (since the pow function doesn't take vectors, we later have to multiply the magnitude of the gradient with the gradient direction)
    vectorField gradientDirection = gradient / (mag(gradient) + SMALL);
    //Info << "grad_dir" << gradientDirection << endl;
    // slip velocity of the last iteration
    vectorField u_wallslip_lastIteration = (*this);

    const label patchI = patch().index();
    scalarField nuw = 1e-6*mag(patch().nf());  
    
    if (db().found("turbulenceModel"))
    {
	const incompressible::turbulenceModel& turbModel =
            db().lookupObject<incompressible::turbulenceModel>
            (
                "turbulenceModel"
            );

	//Info<< "\nTurbulence found\n" << endl;
	nuw = rho_*turbModel.nu()().boundaryField()[patchI];
    }
    else
    {
        nuw = mag(patch().nf());  
	//Info<< "\nTurbulence NOT found\n" << endl;
    }
  
    vectorField u_wallslip = *this;  
    //-------------------------------looping over the facecells starts here----------------------------------

    const fvMesh& mesh     = patch().boundaryMesh().mesh();
   
    const faceList& faces  = mesh.faces();
    const label& startFace = patch().patch().start();
    const label    endFace = startFace + size() - 1 ;
    // Info<< "startFace = " <<  startFace << "endFace = " << endFace << endl;
    label patchID = mesh.boundaryMesh().findPatchID("patchI");
    const polyPatch& cPatch = mesh.boundaryMesh()[patchI];
    label faceId_start = cPatch.start(); 
    forAll(cPatch,faceI)
    {
    vector nHat_face = nHat[faceI];
    //Info << "nHat_face" << nHat_face <<endl;
    vector gradient_face = gradient[faceI];
    //gradient_face = transform(I -sqr(nHat_face),gradient_face);
    vector gradientDirection_face = gradientDirection[faceI];
    scalar nuw_face = nuw[faceI];    
    u_wallslip[faceI] = kh1_*(Foam::sinh(- kh2_*nuw_face*mag(gradient_face)))*gradientDirection_face;
    //Info << "nuw_face" << nuw_face << "walllslip" << u_wallslip[faceI] <<endl;
 
   }       
   // Info<<"mag_grad_face =" << mag(gradient_face) <<"nuw_face =" << nuw_face << endl;
    vectorField::operator=(relaxationFactor_*u_wallslip_lastIteration+(1.0-relaxationFactor_)*u_wallslip);
    Info<<"u_wallslip" <<u_wallslip[50]<<endl;
    fixedValueFvPatchVectorField::updateCoeffs();    
}


// Write
void hatziSlipGenNewtonianRelaxationFvPatchVectorField::write(Ostream& os) const
{
    fvPatchVectorField::write(os);
    os.writeKeyword("rho") << rho_ << token::END_STATEMENT << nl;
    os.writeKeyword("kh2") << kh2_ << token::END_STATEMENT << nl;
    os.writeKeyword("kh1") << kh1_ << token::END_STATEMENT << nl;
    os.writeKeyword("relaxationFactor") << relaxationFactor_ << token::END_STATEMENT << nl;  
    writeEntry("value", os);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

makePatchTypeField(fvPatchVectorField, hatziSlipGenNewtonianRelaxationFvPatchVectorField);

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
